"""
Pyesql is a Python clone of Yesql (https://github.com/krisajenkins/yesql)

Source: https://github.com/KMahoney/pyesql

Differences from Yesql:

- Python does not distinguish between a statement and a query, so there is no need for the `!` suffix.
- Instead of the `?` and `:param` syntax, Pyesql uses the `%(param)s` syntax

"""

from pyesql.parser import parse_file, parse_source, ParseError

__all__ = ['parse_file', 'parse_source', 'ParseError', 'VERSION']

VERSION = '0.0.1'



from unittest import TestCase
from pyesql.parser import LineParser, parse_query, ParseError, parse_queries, parse_source


class TestParseQuery(TestCase):
    def test_simple_query(self):
        query = """
        -- name: test
        QUERY BODY
        """
        name, doc, body = parse_query(LineParser(query))
        self.assertEqual(name, 'test')
        self.assertEqual(doc, '')
        self.assertEqual(body.strip(), 'QUERY BODY')

    def test_multiline_query(self):
        query = """
        -- name: test
        QUERY BODY 1

        QUERY BODY 2

        QUERY BODY 3
        """
        name, doc, body = parse_query(LineParser(query))
        self.assertEqual(name, 'test')
        self.assertEqual(doc, '')
        self.assertEqual([l.strip() for l in body.splitlines()],
                         ['QUERY BODY 1', '', 'QUERY BODY 2', '', 'QUERY BODY 3', ''])

    def test_include_comment_query(self):
        query = """
        -- name: test
        QUERY BODY 1
        -- comment
        QUERY BODY 2
        """
        name, doc, body = parse_query(LineParser(query))
        self.assertEqual(name, 'test')
        self.assertEqual(doc, '')
        self.assertEqual([l.strip() for l in body.splitlines()],
                         ['QUERY BODY 1', '-- comment', 'QUERY BODY 2', ''])

    def test_doc_query(self):
        query = """
        -- name: test
        -- documentation1
        -- documentation2
        QUERY BODY
        """
        name, doc, body = parse_query(LineParser(query))
        self.assertEqual(name, 'test')
        self.assertEqual(doc.strip(), 'documentation1\ndocumentation2')
        self.assertEqual(body.strip(), 'QUERY BODY')

    def test_no_body(self):
        query = """
        -- name: test
        -- documentation1
        -- documentation2
        """
        with self.assertRaises(ParseError):
            parse_query(LineParser(query))

    def test_no_doc_or_body(self):
        query = """
        -- name: test
        """
        with self.assertRaises(ParseError):
            parse_query(LineParser(query))

    def test_empty(self):
        with self.assertRaises(ParseError):
            parse_query(LineParser(""))


class TestParseQueries(TestCase):
    def test_multi_queries(self):
        query = """
        -- name: test1
        -- doc1
        QUERY BODY 1

        -- name: test2
        -- doc2
        QUERY BODY 2
        """
        queries = parse_queries(query)
        self.assertEqual(queries['test1'][0].strip(), 'doc1')
        self.assertEqual(queries['test1'][1].strip(), 'QUERY BODY 1')
        self.assertEqual(queries['test2'][0].strip(), 'doc2')
        self.assertEqual(queries['test2'][1].strip(), 'QUERY BODY 2')

    def test_no_endline(self):
        query = """
        -- name: test1
        -- doc1
        QUERY BODY 1

        -- name: test2
        -- doc2
        QUERY BODY 2"""
        queries = parse_queries(query)
        self.assertEqual(queries['test1'][0].strip(), 'doc1')
        self.assertEqual(queries['test1'][1].strip(), 'QUERY BODY 1')
        self.assertEqual(queries['test2'][0].strip(), 'doc2')
        self.assertEqual(queries['test2'][1].strip(), 'QUERY BODY 2')


class TestObject(TestCase):
    def test_multi_queries(self):
        query = """
        -- name: test1
        -- doc1
        QUERY BODY 1

        -- name: test2
        -- doc2
        QUERY BODY 2
        """
        test_class = parse_source('Test', query)
        test = test_class(None)
        self.assertTrue(hasattr(test, '__init__'))
        self.assertTrue(hasattr(test, 'test1'))
        self.assertTrue(hasattr(test, 'test2'))



import re
from os.path import basename


class ParseError(Exception):
    """An error was encountered in the SQL source"""
    pass


WHITESPACE_RE = re.compile('^\s*$')
NAME_RE = re.compile('^\s*--\s*name:')
DOC_RE = re.compile('^\s*--\s*')


class LineParser(object):
    def __init__(self, source):
        self.lines = iter(source.splitlines())
        self.linecount = 0
        self.next()

    def next(self):
        try:
            self.line = next(self.lines)
        except StopIteration:
            self.line = None
        else:
            self.linecount += 1

    def eof(self):
        return self.line is None

    def skip_whitespace(self):
        while not self.eof() and WHITESPACE_RE.match(self.line):
            self.next()


def parse_query(parser):
    parser.skip_whitespace()

    # Parse name
    if parser.eof() or not NAME_RE.match(parser.line):
        raise ParseError("{}: Expecting '-- name: <query-name>'".format(parser.linecount))

    name = re.sub(NAME_RE, '', parser.line).strip()
    parser.next()
    parser.skip_whitespace()

    if parser.eof():
        raise ParseError('{}: Expecting documentation or query body for query {}'.format(parser.linecount, name))

    # Parse docstring
    doc_lines = []
    while not parser.eof() and DOC_RE.match(parser.line):
        doc_lines.append(re.sub(DOC_RE, '', parser.line))
        parser.next()
    doc = '\n'.join(doc_lines)

    parser.skip_whitespace()

    if parser.eof() or NAME_RE.match(parser.line):
        raise ParseError('{}: Expecting query body for query {}'.format(parser.linecount, name))

    # Parse body
    body_lines = []
    while not parser.eof() and not NAME_RE.match(parser.line):
        body_lines.append(parser.line)
        parser.next()
    body = '\n'.join(body_lines)

    return name, doc, body


def parse_queries(source):
    queries = {}
    parser = LineParser(source)
    while not parser.eof():
        name, doc, body = parse_query(parser)
        queries[name] = (doc, body)
        parser.skip_whitespace()
    return queries


def _make_query_fn(doc, body):
    def inner(self, *args, **kwargs):
        cursor = self.connection.cursor()
        cursor.execute(body, kwargs)
        return cursor.fetchall()
    inner.__doc__ = doc
    return inner


def parse_source(name, source):
    """Parse a SQL source string into a python object.

    Takes a source SQL string and parses it into a new class named
    `name`. The class is initialised with a database connection
    and has a method for each SQL statement.

    Each SQL statement should take the form::

        -- name: <name>
        -- optional documentation
        BODY %(parameter)s

    For example given the SQL source::

        -- name: test1
        -- documentation test1
        SELECT * FROM test

        -- name: test2
        SELECT * FROM test WHERE x = %(y)s

    we can produce this object::

        Example = parse_object('Example', source)
        connection = MySQLdb.connect(...)
        example = Example(connection)
        example.test1()
        example.test2(y=1)

    :param str name: Name of the returned type
    :param str source: Source SQL
    :return: A new Python type
    :raises ParseError: If the Source SQL does not conform

    """
    queries = parse_queries(source)
    obj = {k: _make_query_fn(*v) for k, v in queries.items()}

    def __init__(self, connection):
        self.connection = connection
    obj['__init__'] = __init__

    return type(name, (), obj)


def parse_file(filename, name=None):
    """Parse a SQL source file into a python object.

    Reads `filename` in to memory and returns a new class named after
    the filename. See :py:func:`pyesql.parse_source` for details on the
    returned class.

    :param str filename: Source SQL filename
    :param str name: Optional name of the returned type. Uses filename by default.
    :return: A new Python type
    :raises ParseError: If the Source SQL does not conform

    """
    with open(filename) as f:
        source = f.read()
    return parse_source(name or basename(filename), source)



from setuptools import setup
from pyesql import VERSION

setup(
    name='pyesql',
    version=VERSION,
    description='Python yesql clone',
    url='https://github.com/KMahoney/pyesql',
    packages=['pyesql'],
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Operating System :: OS Independent',
    ]
)
